/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Guillaume Seguin <guillaume@segu.in>
 *         Alina Quereilhac <alina.quereilhac@inria.fr>
 */
#include "tls-data.h"

#if SIMULATOR_THREAD_IMPL

namespace ns3 {

__thread bool g_inRun;

} // namespace ns3

#elif SIMULATOR_PTHREAD_IMPL

#include "ns3/log.h"
#include "ns3/fatal-error.h"

#include <pthread.h>
#include <string.h>

namespace ns3 {

pthread_key_t g_tlsInRunKey;
pthread_once_t g_tlsInRunOnce = PTHREAD_ONCE_INIT;

void
TlsInRunFree (void *tlsInRun)
{
  delete (bool *)tlsInRun;
}

void
TlsInRunInit (void)
{
  int err = pthread_key_create (&g_tlsInRunKey, TlsInRunFree);
  if (err)
    {
      NS_FATAL_ERROR ("pthread_key_create failed:" << err << " =\"" << strerror (err) << "\".");
    }
} 

bool *
TlsGetInRun (void)
{
  // Creation of index key happens once per process
  int err = pthread_once (&g_tlsInRunOnce, TlsInRunInit);
  if (err)
    {
      NS_FATAL_ERROR ("pthread_once failed:" << err << " =\"" << strerror (err) << "\".");
    }

  void *value;
  // The key value should be initialized once per thread.
  if ((value = pthread_getspecific (g_tlsInRunKey)) == NULL)
    {
      value = (void *) new bool ();
      err = pthread_setspecific (g_tlsInRunKey, value);
      if (err)
        {
          NS_FATAL_ERROR ("pthread_setspecific failed:" << err << " =\"" << strerror (err) << "\".");
        }
    }
  return (bool *) value;
}

} // namespace ns3

#endif

#if SIMULATOR_PTHREAD_IMPL

#include "ns3/test.h"

#include <pthread.h>

using namespace ns3;

// ===========================================================================
// Test case to check that the thread run state is correctly isolated between
// threads. 
// ===========================================================================
class TlsInRunTestCase : public TestCase
{
public:
  TlsInRunTestCase ();
  virtual ~TlsInRunTestCase ();
  static bool result;

private:
  virtual void DoSetup (void);
  virtual bool DoRun (void);
  virtual void DoTeardown (void);
  bool DoTestThreadInRun (void);
  static void * TestThreadInRun (void *obj);
};

TlsInRunTestCase::TlsInRunTestCase ()
  : TestCase ("Check that the thread run state is correctly isolated between threads")
{
}

TlsInRunTestCase::~TlsInRunTestCase ()
{
}

void
TlsInRunTestCase::DoSetup (void)
{
}

void
TlsInRunTestCase::DoTeardown (void)
{
}

bool
TlsInRunTestCase::DoRun (void)
{
  NS_TEST_ASSERT_MSG_EQ (TlsIsInRun (), false, "Wrong initial state for TlsIsInRun(). Should be false");
 
  TlsSetInRun (true);
  NS_TEST_ASSERT_MSG_EQ (TlsIsInRun (), true, "Wrong state for TlsIsInRun(). Should be true");

  // Test that the change in the main thread did not affect other threads
  pthread_t thread;
  pthread_create (&thread, NULL, TestThreadInRun, this);
  pthread_join (thread, NULL);

  TlsSetInRun (false);
  NS_TEST_ASSERT_MSG_EQ (TlsIsInRun (), false, "Wrong state for TlsIsInRun(). Should be false");

  return false; 
}

bool
TlsInRunTestCase::DoTestThreadInRun (void)
{
  NS_TEST_ASSERT_MSG_EQ (TlsIsInRun (), false, "Wrong initial state for TlsIsInRun(). Should be false");
  TlsSetInRun (true);
  NS_TEST_ASSERT_MSG_EQ (TlsIsInRun (), true, "Wrong state for TlsIsInRun(). Should be true");

  return false;
}

void *
TlsInRunTestCase::TestThreadInRun (void *obj)
{
  ((TlsInRunTestCase *)obj)->DoTestThreadInRun ();
  return NULL;
}

class TlsDataTestSuite : public TestSuite
{
public:
  TlsDataTestSuite ();
};

TlsDataTestSuite::TlsDataTestSuite ()
  : TestSuite ("tls-data", UNIT)
{
  AddTestCase (new TlsInRunTestCase ());
}

TlsDataTestSuite tlsDataTestSuite;
#endif 


