/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Guillaume Seguin <guillaume@segu.in>
 *         Alina Quereilhac <aquereilhac@inria.fr>
 */

#ifndef TLS_DATA_H
#define TLS_DATA_H

#include "ns3/simulator-config.h"

#include <stdint.h>

namespace ns3 {

/**
 * A per-thread flag is used to mark the thread that is processing the simulation events. 
 * 
 * \returns true if the associated thread is running the simulator.
 */
inline bool TlsIsInRun (void);

/**
 * Set isInRun to true before starting to process simulation events and to false after finishing.
 * 
*/
inline void TlsSetInRun (bool isInRun);

} // namespace ns3

// Each define below represents a different implementation of this API
// These defines are ordered from fastest to slowest on linux
// The slowest implementation is also the most portable: it should
// work safely on cygwin, osx, and all versions of linux.
#if SIMULATOR_THREAD_IMPL

/**
 * This implementation relies on the __thread compiler extension which is
 * a gcc extension available only on linux. There is no osx/cygwin support
 * for this keyword.
 */
namespace ns3 {

// Note: theoretically, we should be able to mark the following variable
// with the initial-exec tls model:  __attribute__((tls_model("initial-exec")))
// However, this does not appear to work for now.
extern __thread bool g_inRun;
inline bool TlsIsInRun (void)
{
  return g_inRun;
}

inline void TlsSetInRun (bool inRun)
{
  g_inRun = inRun;
}

} // namespace ns3

#elif SIMULATOR_PTHREAD_IMPL

/**
 * This implementation relies on standard posix functions only. It is
 * fairly efficient, all things considered
 */

namespace ns3 {

bool * TlsGetInRun (void);

inline bool TlsIsInRun (void)
{
  return *TlsGetInRun ();
}

inline void TlsSetInRun (bool isInRun)
{
  bool *tlsInRun = TlsGetInRun ();
  *tlsInRun = isInRun;
}

} // namespace ns3

#else
#error "unknown TLS implementation"
#endif

#endif /* TLS_DATA_H */
